import argparse
import json
import os.path
import shutil
from pathlib import Path
from subprocess import check_call, CalledProcessError

from hypy_utils import printc, write, json_stringify
from hypy_utils.dict_utils import remove_nones

from .convert_media_types import tgs_to_apng, extract_album_art
from .pyro.consts import HTML

test_text = [
    "test ",
    {"type": "strikethrough", "text": "strikethrough"},
    {"type": "code", "text": "mono"},
    {"type": "italic", "text": "italic"},
    {"type": "underline", "text": "underline"},
    {"type": "bold", "text": "bold"},
    {"type": "spoiler", "text": "spoiler"}
]


def convert_text(text: str | list[dict | str] | None) -> str | None:
    """
    Convert text to markdown/html

    >>> convert_text(test_text)
    'test ~~strikethrough~~ `mono` __italic__ underline **bold** ||spoiler||'

    :param text: Telegram text
    :return: Markdown text
    """
    if text is None:
        return None

    if isinstance(text, str):
        return text

    def convert_entity(e: dict | str) -> str:
        if isinstance(e, str):
            return e
        if isinstance(e, dict):
            t = e["text"]
            match e["type"]:
                case "strikethrough":
                    return f"<del>{t}</del>"
                case "code":
                    return f"<code>{t}</code>"
                case "italic":
                    return f"<em>{t}</em>"
                case "underline":
                    return f"<u>{t}</u>"
                case "bold":
                    return f"<b>{t}</b>"
                case "spoiler":
                    return f'<span class="spoiler"><span>{t}</span></span>'
                case "link":
                    return f'<a href="{t}">{t}</a>'
                # case "mention_name":
                #     id = e['user_id']
                #     return f'<a href="https://t.me/{id}">{t}</a>'
                case "text_link":
                    url = e['href']
                    return f'<a href="{url}">{t}</a>'
                case "hashtag":
                    return f'<a href="{t}">{t}</a>'
                case "mention":
                    url = f'https://t.me/{t.strip("@")}'
                    return f'<a href="{url}">{t}</a>'
                case "custom_emoji":
                    url: str = e['document_id']
                    # if url.endswith(".webm"):
                    #     url = webm_to_apng(url)
                    return f'<i class="custom-emoji" emoji-src="{url}">{t}</i>'
                case _:
                    return t

    text = [[e, convert_entity(e)] for e in text]

    return "".join([t[1] for t in text])


def plain_text(text: str | list[dict | str] | None) -> str | None:
    """
    Convert text, remove all formats

    :param text: Telegram export text object
    :return: Plain text
    """
    if text is None:
        return None

    if isinstance(text, str):
        return text

    acc = ""
    for t in text:
        if isinstance(t, str):
            acc += t
        if isinstance(t, dict):
            acc += t["text"]

    return acc


def process_file_path(path: str | None) -> str | None:
    """
    Ensure file paths are url-safe
    """
    if path is None:
        return None

    # Convert tgs stickers to apng
    if path.endswith(".tgs"):
        path = str(tgs_to_apng(p / path).relative_to(p))

    return path


def parse_file(d: dict) -> dict | None:
    file = d.get("file")
    if file is None:
        return None

    # Create file
    file = {
        "url": process_file_path(file),
        "thumb": process_file_path(d.get("thumbnail")),
        "mime_type": d.get("mime_type"),
        "size": os.path.getsize(p / file),
        "original_name": d.get("original_name"),

        # Media
        "media_type": d.get("media_type"),

        # Video/audio/gif
        "duration": d.get("duration_seconds"),

        # Video/sticker/gif
        "width": d.get("width"),
        "height": d.get("height"),

        # Sticker
        "sticker_emoji": d.get("sticker_emoji"),

        # Audio
        "title": d.get("title"),
        "performer": d.get("performer")
    }

    # Convert image file to photo
    if not file['media_type'] and file['mime_type'].startswith('image'):
        file['media_type'] = "photo"

    # Add image for missing album cover art
    if file['media_type'] == 'audio_file' and not file['thumb']:
        fp = p / file['url']
        op = extract_album_art(fp)
        if op:
            file['thumb'] = op.relative_to(p)

    return file


def get_image(d: dict) -> dict:
    return {"url": d.get("photo"), "width": d.get("width"), "height": d.get("height")}


def convert_msg(d: dict) -> dict | None:
    """
    Convert a message object

    :param d: Message object dict from telegram export
    :return: Message object for tg-blog model
    """
    # Message group quirks
    grp = d.get("media_group_id")
    if grp in processed_groups:
        return None
    if grp is not None:
        processed_groups[grp] = d['id']

    def get_group_text():
        if grp is None:
            return convert_text(d.get("text"))
        for msg in groups[grp]:
            t = convert_text(msg.get("text"))
            if t:
                return t
        return None

    def get_group_images():
        if d.get("photo") is None:
            return None
        if grp is None:
            return [get_image(d)]
        return [get_image(m) for m in groups[grp]]

    def get_group_files():
        if d.get("file") is None:
            return None
        if grp is None:
            return [parse_file(d)]
        return [parse_file(m) for m in groups[grp]]

    reply = id_map.get(d.get("reply_to_message_id"))
    # Resolve reply group to the final message id
    if reply and reply.get('media_group_id'):
        reply = id_map.get(processed_groups[reply.get("media_group_id")])

    msg = {
        "id": d["id"],
        "date": d["date"],
        "type": None if d.get("type") == "message" else d.get("type"),
        "text": get_group_text(),
        "views": d.get("views"),  # Views cannot be exported in the current version
        "images": get_group_images(),
        "forwarded_from": d.get("forwarded_from"),

        # TODO: Add this in front end
        "video": None if d.get("media_type") != "video_file" else {
            "thumb": d.get("thumbnail"), "duration": d.get("duration_seconds"), "src": d.get("file")
        },
        "reply": None if reply is None else {
            "id": reply['id'],
            "text": plain_text(reply.get("text")),
            "thumb": reply.get("thumbnail") or reply.get("photo"),
        },
        "author": d.get("author"),
        "files": get_group_files()
        # TODO: Add more fields
    }

    # Convert image file group to photo group
    if msg.get('files') and msg['files'][0]['media_type'] == 'photo':
        msg['images'] = msg.pop('files')

    return remove_nones(msg)


def infer_groups(msgs: list[dict]):
    """
    Infer message media/file/photo groups from timestamp and media type

    :param msgs: Messages (will be modified)
    :return: None
    """
    i = 0
    c_group = 0
    c_type = ""
    c_time = 0
    c_count = 0
    while i < len(msgs):
        it = msgs[i]
        time = int(it['date_unixtime'])

        if it.get('photo'):
            ty = "photo"
        else:
            ty = it.get("media_type")

        try:
            # Type cannot be a sticker / video / regular message
            assert ty is not None and ty != "sticker" and ty != 'video'

            # Timestamps are within 5 seconds
            assert abs(c_time - time) < 5

            # Types must match
            assert c_type == ty

            # Assign group id
            msgs[i - 1]['media_group_id'] = c_group
            it['media_group_id'] = c_group
            c_count += 1

        except AssertionError:
            c_type = ty
            c_time = time
            c_count = 0
            c_group += 1
            pass

        i += 1


p: Path
id_map: dict[int, dict]
groups: dict[int, list[dict]]
processed_groups: dict[int, int]


def convert_original_filenames(json_path: Path):
    """
    Convert filenames in the original result.json
    """
    renamed: dict[str, str] = {}
    results = json.loads(json_path.read_text())

    if results.get('processed'):
        return

    for k in ['file', 'thumbnail']:
        for r in [r for r in results['messages'] if k in r]:
            orig = str(r[k])
            if orig in renamed:
                r[k] = renamed[orig]
                continue

            f = p / r[k]
            new = f.with_name(f"{r['id']}{'_thumb' if k == 'thumbnail' else ''}{f.suffix}")
            np = str(new.relative_to(p))
            if new != f:
                printc(f"&6Renaming &r{orig} &6to &r{np}")
                shutil.move(f, new)
                if k == 'file':
                    r['original_name'] = f.name
                r[k] = np
                renamed[orig] = np
    results['processed'] = True

    write(json_path, json_stringify(results, indent=2))


def run():
    global p, id_map, groups, processed_groups
    parser = argparse.ArgumentParser("Telegram export converter",
                                     description="A tool to convert exported json into tg-blog json")
    parser.add_argument("dir", help="Export directory")
    args = parser.parse_args()

    p = Path(args.dir)
    f = p / "result.json"
    assert f.is_file(), f"Error: File {f} not found"

    # Convert file names to url-safe file names
    convert_original_filenames(f)

    # Read export result json
    j: list[dict] = json.loads(f.read_text())["messages"]
    id_map = {d['id']: d for d in j}

    # Assign groups
    infer_groups(j)

    # Group groups
    tmp_grouped: list[dict] = [d for d in j if 'media_group_id' in d]
    group_ids: set[int] = {d['media_group_id'] for d in tmp_grouped}
    groups = {g: [d for d in tmp_grouped if d['media_group_id'] == g] for g in group_ids}
    processed_groups = {}

    # print(json.dumps(j, indent=2, ensure_ascii=False))

    # Convert
    j = [convert_msg(d) for d in j]
    j = [d for d in j if d is not None]

    write(p / "posts.json", json_stringify(j, indent=2))
    write(p / "index.html", HTML.replace("$$POSTS_DATA$$", json_stringify(j)))

    printc(f"&aDone! Saved to {p / 'posts.json'}")


if __name__ == '__main__':
    run()



import os
import zlib
from pathlib import Path
from shutil import which
from subprocess import check_call, CalledProcessError, check_output

from hypy_utils import printc


SCRIPT_PATH = Path(__file__).parent

NODE_BIN_PATHS: list[Path] = [
    SCRIPT_PATH / 'node_modules/.bin',
    Path('node_modules/.bin'),
    Path.home() / 'node_modules/.bin',
    Path.home() / '.config/yarn/global/node_modules/.bin',
    Path('/usr/local/bin'),
    Path('/usr/bin')
]


def find_node_bin(name: str, pkg_name: str) -> Path:
    # Find bin path
    path = which(name)
    if path and os.path.isfile(path):
        return Path(path)

    for bin in NODE_BIN_PATHS:
        if (bin / name).is_file():
            return bin / name

    printc(f"&cError! Cannot find node executable {name}. \n"
           f"Make sure to install it using 'yarn global add {pkg_name}'")
    exit(3)


def tgs_to_apng(tgs: str | Path) -> Path:
    """
    Convert .tgs vector animation into .apng bitmap animation

    :param tgs: TGS file
    :return: Converted path (str)
    """
    tgs = Path(tgs)
    out = tgs.with_suffix(".apng")

    if not out.is_file():
        # Decompress to json
        json = tgs.with_suffix(".json")
        json.write_bytes(zlib.decompress(tgs.read_bytes(), 15 + 32))

        # Convert json to apng
        check_call([find_node_bin("puppeteer-lottie", "puppeteer-lottie-cli"), '-i', json, '-o', out])

        # Delete json
        os.remove(json)

    return out


def webm_to_apng(webm: str, p: Path) -> str:
    """
    Convert .webm contained animation into .apng bitmap animation

    :param webm: Webm file
    :param p: Output path
    :return: Converted path (str)
    """
    out = str(Path(webm).with_suffix(".apng"))

    if not (p / out).is_file():
        # Convert webm to apng using ffmpeg
        cmd = ['ffmpeg', '-c:v', 'libvpx-vp9',
               # '-pix_fmt', 'yuva420p',
               '-i', str(p / webm),
               '-plays', '0',
               str(p / out)]
        print(' '.join(cmd))
        check_call(cmd)

    return out


def extract_album_art(f: Path) -> Path | None:
    f = Path(f)
    op = f.with_name(f.stem + '_thumb.png')
    if op.is_file():
        return op

    try:
        check_output(['ffmpeg', '-y', '-i', f, '-an', '-vf', 'scale=250:-1', op])
        printc(f"&a> Success! Saved to {op}")
        return op
    except CalledProcessError as e:
        printc(f"&c> Failed to extract album art: {e}")



from hypy_utils.dict_utils import remove_nones


def group_msgs(msgs: list[dict]) -> list[dict]:
    """
    Merge message files into groups

    Conditions: Messages will be sorted by increasing ID (with the oldest message on top)

    :param msgs: Messages
    :return: Messages with file groups
    """
    # Sort messages
    msgs = sorted(msgs, key=lambda x: x['id'])

    # Find groups
    id_map = {m['id']: m for m in msgs}
    tmp_grouped: list[dict] = [d for d in msgs if 'media_group_id' in d]
    gids: set[int] = {d['media_group_id'] for d in tmp_grouped}
    groups: dict[int, list[dict]] = {g: [d for d in tmp_grouped if d['media_group_id'] == g] for g in gids}

    # Group messages
    result = []
    reply_map: dict[int, dict] = {}
    for gid, grp in groups.items():
        # Loop through each message, find the first message with text
        def find_dominant():
            for m in grp:
                if 'text' in m:
                    return m
            return grp[0]

        m = find_dominant()
        result.append(m)
        reply_map.update({a['id']: m for a in grp})

        # Group files & images into a list
        m['files'] = [a.get('file') for a in grp if 'file' in a] or None
        m['images'] = [a.get('image') for a in grp if 'image' in a] or None

        # Clean up nones
        m.pop('file', None)
        m.pop('image', None)
        if not m['files']:
            del m['files']
        if not m['images']:
            del m['images']

    # Change all reply references
    for m in msgs:
        if 'reply_id' not in m:
            continue
        rid = m.pop('reply_id')
        rm = reply_map.get(rid) or id_map[rid]

        m['reply'] = {
            'id': rm['id'],
            'text': rm.get('text'),
            'thumb': ((rm.get('files') or rm.get('images') or [None])[0] or {}).get('thumb')
        }

    # Add non-grouped messages to results
    grouped_ids = {m['id'] for v in groups.values() for m in v}
    ungrouped = [m for m in msgs if m['id'] not in grouped_ids]
    for m in ungrouped:
        if 'file' in m:
            m['files'] = [m.pop('file')]
        if 'image' in m:
            m['images'] = [m.pop('image')] or None
    result += ungrouped

    return sorted(remove_nones(result), key=lambda x: x['id'])



import os
from dataclasses import dataclass
from pathlib import Path

import toml
from hypy_utils import printc


@dataclass
class Config:
    api_id: int
    api_hash: str
    bot_token: str
    exports: list[dict]


def load_config(path: str = "config.toml") -> Config:
    if os.getenv('tgc_config'):
        return Config(**toml.loads(os.getenv('tgc_config')))

    fp = os.getenv('tgc_config_path')

    if fp is None or not os.path.isfile(fp):
        fp = path
    if fp is None or not os.path.isfile(fp):
        fp = Path.home() / ".config" / "tgc" / "config.toml"

    fp = Path(fp)

    if not fp.is_file():
        printc(f"&cConfig file not found in either {path} or {fp} \nPlease put your configuration in the path")
        exit(3)

    return Config(**toml.loads(fp.read_text()))



from pathlib import Path

from pyrogram.enums import MessageMediaType

MEDIA_TYPE_MAP: dict[MessageMediaType, str] = {
    MessageMediaType.PHOTO: "photo",
    MessageMediaType.STICKER: "sticker",
    MessageMediaType.VOICE: "voice_message",
    MessageMediaType.AUDIO: "audio_file",
    MessageMediaType.ANIMATION: "animation",
    MessageMediaType.VIDEO: "video_file",
    MessageMediaType.VIDEO_NOTE: "video_file",
    MessageMediaType.DOCUMENT: None,

    # TODO: Support these in web ui
    MessageMediaType.CONTACT: "contact",
    MessageMediaType.POLL: "poll",
    MessageMediaType.WEB_PAGE: "web_page",
    MessageMediaType.LOCATION: "location",
    MessageMediaType.VENUE: "location"
}

HTML = (Path(__file__).parent.parent / "tg-blog.html").read_text()



import argparse
import asyncio
from pathlib import Path

import uvloop
from PIL import Image
from hypy_utils import printc, json_stringify, write
from hypy_utils.dict_utils import remove_keys
from pyrogram import Client
from pyrogram.file_id import FileId
from pyrogram.types import User, Chat, Message, Sticker

from .config import load_config, Config
from .consts import HTML
from .convert import convert_text, convert_media_dict
from .download_media import download_media, has_media, guess_ext, download_media_urlsafe
from .grouper import group_msgs
from ..convert_export import remove_nones
from ..convert_media_types import tgs_to_apng

uvloop.install()


def effective_text(msg: Message) -> str:
    """
    Get effective text of a message in HTML
    """
    if msg.text:
        return convert_text(msg.text, msg.text.entities)
    if msg.caption:
        return convert_text(msg.caption, msg.caption.entities)
    if msg.service:
        return str(msg.service).split(".")[-1].replace("_", " ").capitalize()


def _download_media_helper(args: list) -> Path:
    return asyncio.run(download_media(app, *args))


def get_user_name(user: User) -> str:
    name = user.first_name or ""
    if user.last_name:
        name += " " + user.last_name
    return name


async def process_message(msg: Message, path: Path, export: dict) -> dict:
    media_path = path / "media"

    m = {
        "id": msg.id,
        "date": msg.date,
        "type": 'service' if msg.service else None,
        "text": effective_text(msg),
        "author": msg.author_signature,
        "views": msg.views,
        "forwards": msg.forwards,
        "forwarded_from": {
            "name": get_user_name(msg.forward_from),
            "url": f'https://t.me/{msg.forward_from.username}' if msg.forward_from.username else None,
        } if msg.forward_from else None,
        "media_group_id": msg.media_group_id,
        "reply_id": msg.reply_to_message_id,
        "file": convert_media_dict(msg)
    }

    # Download file
    f = m.get('file')

    async def dl_media():
        fp, name = await download_media_urlsafe(app, msg, directory=media_path,
                                                max_file_size=int((export.get('size_limit_mb') or 0) * 1000_000))
        if fp is None:
            return
        f['original_name'] = name

        # Convert tgs sticker
        if fp.suffix == '.tgs':
            fp = Path(tgs_to_apng(fp))

        f['url'] = str(fp.absolute().relative_to(path.absolute()))
        f['size'] = f.pop('file_size', None)

        # Download the largest thumbnail
        if f.get('thumbs'):
            thumb: dict = max(f['thumbs'], key=lambda x: x['file_size'])
            ext = guess_ext(app, FileId.decode(thumb['file_id']).file_type, None)
            fp = await download_media(app, thumb['file_id'], directory=media_path,
                                      fname=fp.with_suffix(fp.suffix + f'_thumb{ext}').name)
            f['thumb'] = str(fp.absolute().relative_to(path.absolute()))
            del f['thumbs']

    if has_media(msg):
        await dl_media()

    # Move photo to its own key
    if f:
        mt = f.get('media_type')
        if mt == 'photo' or (not mt and (f.get('mime_type') or "").startswith("image")):
            img = m['image'] = m.pop('file')

            # Read image size
            img['width'], img['height'] = Image.open(path / img['url']).size

    return remove_keys(remove_nones(m), {'file_id', 'file_unique_id'})


async def download_custom_emojis(msgs: list[Message], results: list[dict], path: Path):
    # List custom emoji ids
    ids = list({e.custom_emoji_id for msg in msgs if msg.text and msg.text.entities for e in msg.text.entities if e.custom_emoji_id})
    orig_ids = list(ids)

    # Query stickers 200 ids at a time
    stickers: list[Sticker] = []
    while ids:
        stickers += await app.get_custom_emoji_stickers(ids[:200])
        ids = ids[200:]

    # Download stickers
    for id, s in zip(orig_ids, stickers):
        ext = guess_ext(app, FileId.decode(s.file_id).file_type, s.mime_type)
        op = (await download_media(app, s, path / "emoji", f'{id}{ext}')).absolute().relative_to(path.absolute())

        # Replace sticker paths
        for r in results:
            if "text" in r:
                r['text'] = r['text'].replace(f'<i class="custom-emoji" emoji-src="emoji/{id}">', f'<i class="custom-emoji" emoji-src="{op}">{s.emoji}')


async def process_chat(chat_id: int, path: Path, export: dict):
    chat: Chat = await app.get_chat(chat_id)
    printc(f"&aChat obtained. Chat name: {chat.title} | Type: {chat.type} | ID: {chat.id}")

    # Crawl 200 messages each request
    print("Crawling channel posts...")
    msgs = []
    for i in range(999):
        start_idx = i * 200 + 1
        end_idx = start_idx + 200

        additional_msgs = await app.get_messages(chat.id, range(start_idx, end_idx))
        additional_msgs = [m for m in additional_msgs if not m.empty]
        msgs += additional_msgs
        print(f"> {len(msgs)} total messages... (up to ID #{end_idx - 1})")

        if not additional_msgs:
            print("> All 200 messages are empty, we're done.")
            break

    # print(msgs)
    results = [await process_message(m, path, export) for m in msgs]
    await download_custom_emojis(msgs, results, path)

    # Group messages
    results = group_msgs(results)

    write(path / "posts.json", json_stringify(results, indent=2))
    write(path / "index.html", HTML.replace("$$POSTS_DATA$$", json_stringify(results)))

    printc(f"&aDone! Saved to {path / 'posts.json'}")


async def run_app():
    me: User = await app.get_me()
    printc(f"&aLogin success! ID: {me.id} | is_bot: {me.is_bot}")
    for export in cfg.exports:
        await process_chat(int(export["chat_id"]), Path(export["path"]), export)


cfg: Config
app: Client


def run():
    global app, cfg
    parser = argparse.ArgumentParser("Telegram Channel Message to Public API Crawler")
    parser.add_argument("config", help="Config path", nargs="?", default="config.toml")
    args = parser.parse_args()
    cfg = load_config(args.config)

    app = Client("Bot", cfg.api_id or 2048, cfg.api_hash or "b18441a1ff607e10a989891a5462e627",
                 **(dict(bot_token=cfg.bot_token) if cfg.bot_token else {}))

    with app:
        asyncio.get_event_loop().run_until_complete(run_app())



#  Pyrogram - Telegram MTProto API Client Library for Python
#  Copyright (C) 2017-present Dan <https://github.com/delivrance>
#
#  This file is part of Pyrogram.
#
#  Pyrogram is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Pyrogram is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Pyrogram.  If not, see <http://www.gnu.org/licenses/>.
import shutil
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Callable, Any

from hypy_utils import ensure_dir, md5
from hypy_utils.file_utils import escape_filename
from pyrogram import types, Client
from pyrogram.file_id import FileId, FileType, PHOTO_TYPES
from pyrogram.types import Message


def guess_ext(client: Client, file_type: int, mime_type: str | None) -> str:
    guessed_extension = client.guess_extension(mime_type) if mime_type else None

    if file_type in PHOTO_TYPES:
        return ".jpg"
    elif file_type == FileType.VOICE:
        return guessed_extension or ".ogg"
    elif file_type in (FileType.VIDEO, FileType.ANIMATION, FileType.VIDEO_NOTE):
        return guessed_extension or ".mp4"
    elif file_type == FileType.DOCUMENT:
        return guessed_extension or ".zip"
    elif file_type == FileType.STICKER:
        return guessed_extension or ".webp"
    elif file_type == FileType.AUDIO:
        return guessed_extension or ".mp3"
    else:
        return ".unknown"


def has_media(message: Message) -> Any | None:
    available_media = ("audio", "document", "photo", "sticker", "animation", "video", "voice", "video_note",
                       "new_chat_photo")

    if isinstance(message, types.Message):
        for kind in available_media:
            media = getattr(message, kind, None)

            if media is not None:
                break
        else:
            return None
    else:
        media = message

    return media


def get_file_name(client: Client, message: Message) -> tuple[str, FileId]:
    """
    Guess a file name of a message media

    :param client: Client
    :param message: Message or media
    :return: File name, file id object
    """
    media = has_media(message)

    if isinstance(media, str):
        file_id_str = media
    else:
        file_id_str = media.file_id

    file_id_obj = FileId.decode(file_id_str)
    file_type = file_id_obj.file_type

    mime_type = getattr(media, "mime_type", "")
    date = getattr(media, "date", None)

    file_name = getattr(media, "file_name", None)

    if not file_name:
        file_name = f"{FileType(file_type).name.lower()}"
        if date:
            file_name += f"_{date.strftime('%Y-%m-%d_%H-%M-%S')}"
        file_name += guess_ext(client, file_type, mime_type)
    file_name = escape_filename(file_name)

    return file_name, file_id_obj


async def download_media(
        client: Client,
        message: types.Message,
        directory: str | Path = "media",
        fname: str | None = None,
        progress: Callable = None,
        progress_args: tuple = (),
        max_file_size: int = 0
) -> Path | Path:
    directory: Path = ensure_dir(directory)

    media = has_media(message)

    # Check filesize
    fsize = getattr(media, 'file_size', 0)
    if max_file_size and fsize > max_file_size:
        print(f"Skipped {fname} because of file size limit ({fsize} > {max_file_size})")
        return None

    file_name, file_id_obj = get_file_name(client, message)
    file_name = fname or file_name

    p = directory / file_name
    if p.exists():
        return p

    print(f"Downloading {p.name}...")

    return Path(await client.handle_download(
        (file_id_obj, directory, file_name, False, fsize, progress, progress_args)
    ))


async def download_media_urlsafe(
        client: Client,
        message: types.Message,
        directory: str | Path = "media",
        fname: str | None = None,
        progress: Callable = None,
        progress_args: tuple = (),
        max_file_size: int = 0
) -> tuple[Path, str]:
    """
    Download media into a renamed file

    :return: Renamed file path, original file name
    """
    file_name, file_id_obj = get_file_name(client, message)
    renamed = str(message.id) + Path(file_name).suffix
    return await download_media(client, message, directory, renamed, progress, progress_args, max_file_size), file_name



from hypy_utils.dict_utils import deep_dict
from pyrogram.enums import MessageEntityType
from pyrogram.parser import utils
from pyrogram.types import MessageEntity, Message

from tgc.pyro.consts import MEDIA_TYPE_MAP


def convert_media_dict(msg: Message) -> dict:
    def helper():
        for f in ['photo', 'video', 'audio', 'voice', 'document', 'sticker', 'animation', 'video_note', 'contact',
                  'location', 'venue', 'poll', 'web_page']:
            dct = getattr(msg, f, None)
            if dct:
                return dict(vars(dct))
        return {}

    d = deep_dict(helper(), {'_client'})
    if d:
        d['media_type'] = MEDIA_TYPE_MAP.get(msg.media)

    # Move location to one place
    if msg.venue:
        d.pop('location', None)
        if msg.venue.location:
            d['longitude'] = msg.venue.location.longitude
            d['latitude'] = msg.venue.location.latitude

    return d


def entity_start_end(text: str, en: MessageEntity) -> tuple[str, str] | None:
    """
    Convert a message entity to a start tag and an end tag for HTML
    """
    text = text[en.offset:en.offset + en.length]
    match en.type:
        case MessageEntityType.STRIKETHROUGH:
            return "<del>", "</del>"
        case MessageEntityType.CODE:
            return "<code>", "</code>"
        case MessageEntityType.ITALIC:
            return "<em>", "</em>"
        case MessageEntityType.UNDERLINE:
            return "<u>", "</u>"
        case MessageEntityType.BOLD:
            return "<b>", "</b>"
        case MessageEntityType.BLOCKQUOTE:
            return "<blockquote>", "</blockquote>"
        case MessageEntityType.SPOILER:
            return '<span class="spoiler"><span>', '</span></span>'
        case MessageEntityType.TEXT_LINK:
            return f'<a href="{en.url}">', '</a>'
        case MessageEntityType.URL:
            return f'<a href="{text}">', '</a>'
        case MessageEntityType.HASHTAG:
            return f'<a href="#{text}">', '</a>'
        case MessageEntityType.MENTION:
            return f'<a href="https://t.me/{text.strip("@")}">', '</a>'
        case MessageEntityType.CUSTOM_EMOJI:
            # This will be replaced later
            return f'<i class="custom-emoji" emoji-src="emoji/{en.custom_emoji_id}">', '</i>'
        case MessageEntityType.PRE:
            lang = en.language
            return f'<pre language="{lang}">' if lang else f"<pre>", f"</pre>"
        case _:
            return None


def convert_text(text: str, entities: list[MessageEntity]) -> str:
    """
    Convert text to HTML
    """
    text = utils.add_surrogates(text)

    entities_offsets = []

    for entity in entities:
        start = entity.offset
        end = start + entity.length

        tags = entity_start_end(text, entity)
        if tags is None:
            continue

        entities_offsets.append((tags[0], start,))
        entities_offsets.append((tags[1], end,))

    entities_offsets = map(
        lambda x: x[1],
        sorted(
            enumerate(entities_offsets),
            key=lambda x: (x[1][1], x[0]),
            reverse=True
        )
    )

    for entity, offset in entities_offsets:
        text = text[:offset] + entity + text[offset:]

    return utils.remove_surrogates(text)
